/*
	dcc-cmd
	
	A simple DCC command station using the dcc-kiss interface.
	
	Copyright Jeroen Vreeken (jeroen@vreeken.net), 2010, 2011, 2012, 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <fcntl.h>
#include <limits.h>
#include <math.h>
#include <pthread.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <termios.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>

#include "dcc_cmd.h"
#include "dcc_srcpd.h"
#include "dcc_interface.h"
#include "dcc_decoder.h"

#define SPEEDSTEP 0.025

double round(double val)
{
	if (val >= 0.0)
		return (int)(val + 0.5);
	else
		return (int)(val - 0.5);
}


struct dcc_accessory_decoder *dcc_accessory_decoder_list = NULL;
pthread_mutex_t dcc_accessory_decoder_list_mutex = PTHREAD_MUTEX_INITIALIZER;

static int fd_dcc;
static bool cur_power = false;
bool rail_power = false;



int dcc_speed28_send(int fd, struct dcc_address *address, double speed)
{
	unsigned char dcc_packet[1];
	bool direction, bitc;
	int speedi;

	direction = !signbit(speed);

	speedi = abs(round(speed * 28));
	
	if (speedi & 1 || speedi == 0) {
		bitc = false;
	} else {
		bitc = true;
	}
	
	dcc_packet[0] = 0x40 | 
	     (direction << 5) |
	     (bitc << 4) |
	     ((speedi >> 1) + (speedi ? 1 : 0));
	
	return dcc_send(fd, dcc_packet, 1, address, false);
}

int dcc_speed128_send(int fd, struct dcc_address *address, double speed)
{
	unsigned char dcc_packet[2];
	bool direction;
	int speedi;

	direction = !signbit(speed);

	speedi = abs(round(speed * 126));
	
	dcc_packet[0] = 0x3f;
	dcc_packet[1] = 
	     (direction << 7) |
	     (speedi + (speedi ? 1 : 0));

	return dcc_send(fd, dcc_packet, 2, address, false);
}

int dcc_cv_set(struct dcc_address *address, int cv, unsigned char val)
{
	unsigned char dcc_packet[3];
	int ret, i;
	
	dcc_send_lock();

	cv-=1;

	for (i = 0; i < 10; i++) {
		struct timespec ts;

		ts.tv_sec = 0;
		ts.tv_nsec = 100000000;
		nanosleep(&ts, NULL);
	
		dcc_packet[0] = 0xec | (cv >> 8);
		dcc_packet[1] = cv & 0xff;
		dcc_packet[2] = val;
	
		ret = dcc_send_real(fd_dcc, dcc_packet, 3, address, true);
		if (ret != 3) {
			printf("Could not send CV packet: %d\n", ret);
			break;
		} else {
			printf("Sent CV packet %d\n", i);
		}
	}
	
	dcc_send_unlock();
	
	return ret;
}

static int dcc_functions_send(int fd, struct dcc_decoder *decoder, bool force)
{
	unsigned char dcc_packet[2];
	int functions = decoder->functions;
	struct dcc_address *address = &decoder->address;

	switch(decoder->nr_functions) {
		case 29:
		case 28:
		case 27:
		case 26:
		case 25:
		case 24:
		case 23:
		case 22:
			if (force || 
			    DCC_FUNCTION_21_28(decoder->func_changed)) {
				dcc_packet[0] = 0xc0 | 0x1f;
				dcc_packet[1] = DCC_FUNCTION_21_28(functions);
				dcc_send(fd, dcc_packet, 2, address, false);
			}
		case 21:
		case 20:
		case 19:
		case 18:
		case 17:
		case 16:
		case 15:
		case 14:
			if (force ||
			    DCC_FUNCTION_13_20(decoder->func_changed)) {
				dcc_packet[0] = 0xc0 | 0x1e;
				dcc_packet[1] = DCC_FUNCTION_13_20(functions);
				dcc_send(fd, dcc_packet, 2, address, false);
			}
		case 13:
		case 12:
		case 11:
		case 10:
			if (force ||
			    DCC_FUNCTION_GROUP2_H(decoder->func_changed)) {
				dcc_packet[0] = 0xa0 | DCC_FUNCTION_GROUP2_H(functions);
				dcc_send(fd, dcc_packet, 1, address, false);
			}
		case 9:
		case 8:
		case 7:
		case 6:
			if (force ||
			    DCC_FUNCTION_GROUP2_L(decoder->func_changed)) {
				dcc_packet[0] = 0xb0 | DCC_FUNCTION_GROUP2_L(functions);
				dcc_send(fd, dcc_packet, 1, address, false);
			}
		case 5:
		case 4:
		case 3:
		case 2:
		case 1:
			if (force ||
			    DCC_FUNCTION_GROUP1(decoder->func_changed)) {
				dcc_packet[0] = 0x80 | DCC_FUNCTION_GROUP1(functions);
				dcc_send(fd, dcc_packet, 1, address, false);
			}
		default:
			break;
	}
	
	return 0;
}

int dcc_accessory_decoder_send(int fd, struct dcc_accessory_decoder *decoder)
{
	unsigned char dcc_packet[2];
	int value = decoder->value;
	struct dcc_address *address = &decoder->address;
	bool val;
	
	if (address->type == DCC_ADDRESS_EXTENDED_ACCESSORY) {
		printf("Extended accessories are not yet supported\n");
		return -1;
	}
	
	val = value & 1;
	dcc_packet[0] = 0x80 | (val << 3);
	dcc_send(fd, dcc_packet, 1, address, false);

	val = value & 2;
	dcc_packet[0] = 0x81 | (val << 3);
	dcc_send(fd, dcc_packet, 1, address, false);

	return 0;
}



void dcc_decoder_function_toggle(struct dcc_decoder *decoder, int function)
{
	decoder->functions ^= function;
	printf("Decoder %d (@%p) f: 0x%x\n", decoder->address.address, decoder, decoder->functions);
}




struct dcc_accessory_decoder *dcc_accessory_decoder_add(
    enum dcc_address_type type, 
    int address)
{
	struct dcc_accessory_decoder *decoder;
	
	decoder = malloc(sizeof(struct dcc_accessory_decoder));
	if (!decoder) {
		perror("Could not alloc accessory decoder structure\n");
		return NULL;
	}
	
	decoder->address.type = type;
	decoder->address.address = address;
	printf("adding accessory decoder with address %d\n",
	    decoder->address.address);
	decoder->value = 0;
	decoder->value_t = 0;
	decoder->lock = 0;
	
	pthread_mutex_lock(&dcc_accessory_decoder_list_mutex);
	decoder->refcount = 2;
	decoder->next = dcc_accessory_decoder_list;
	dcc_accessory_decoder_list = decoder;
	pthread_mutex_unlock(&dcc_accessory_decoder_list_mutex);
	
	return decoder;
}


static void dcc_accessory_decoder_unlist(struct dcc_accessory_decoder *decoder)
{
	struct dcc_accessory_decoder **listp;
	
	for (listp = &dcc_accessory_decoder_list; *listp; listp = &(*listp)->next) {
		if (*listp == decoder) {
			*listp = decoder->next;
			break;
		}
	}
	decoder->next = NULL;
}


struct dcc_accessory_decoder *dcc_accessory_decoder_find(int address)
{
	struct dcc_accessory_decoder *decoder = NULL;
	
	pthread_mutex_lock(&dcc_accessory_decoder_list_mutex);
	for (decoder = dcc_accessory_decoder_list; decoder; decoder = decoder->next) {
		if (decoder->address.address == address) {
			decoder->refcount++;
			break;
		}
	}
	pthread_mutex_unlock(&dcc_accessory_decoder_list_mutex);
	return decoder;
}

void dcc_accessory_decoder_output_set(struct dcc_accessory_decoder *decoder,
    int port, bool value)
{
	if (value) {
		decoder->value |= (1 << port);
	} else {
		decoder->value &= ~(1 << port);
	}
	decoder->value_t = time(NULL);
}

void dcc_accessory_decoder_put(struct dcc_accessory_decoder *decoder)
{
	pthread_mutex_lock(&dcc_accessory_decoder_list_mutex);
	decoder->refcount--;
	if (decoder->refcount == 0) {
		dcc_accessory_decoder_unlist(decoder);
		free(decoder);
	}
	pthread_mutex_unlock(&dcc_accessory_decoder_list_mutex);
}

void dcc_accessory_decoder_del(struct dcc_accessory_decoder *decoder)
{
	pthread_mutex_lock(&dcc_accessory_decoder_list_mutex);
	dcc_accessory_decoder_unlist(decoder);
	pthread_mutex_unlock(&dcc_accessory_decoder_list_mutex);
	dcc_accessory_decoder_put(decoder);
}

struct dcc_accessory_decoder *dcc_accessory_decoder_loop(
    struct dcc_accessory_decoder *decoder)
{
	struct dcc_accessory_decoder *next = NULL;

	if (!decoder) {
		pthread_mutex_lock(&dcc_accessory_decoder_list_mutex);
		next = dcc_accessory_decoder_list;
		if (next) {
			next->refcount++;
		}
		pthread_mutex_unlock(&dcc_accessory_decoder_list_mutex);
	} else {
		pthread_mutex_lock(&dcc_accessory_decoder_list_mutex);
		next = decoder->next;
		if (next) {
			next->refcount++;
		}
		pthread_mutex_unlock(&dcc_accessory_decoder_list_mutex);
		
		dcc_accessory_decoder_put(decoder);
	}

	return next;
}




bool dcc_update(int fd, bool speed_force, bool func_force, bool accessory_force)
{
	struct dcc_decoder *decoder = NULL;
	struct dcc_accessory_decoder *acc = NULL;
	bool update_done = false;
	time_t t = time(NULL);

	while ((decoder = dcc_decoder_loop(decoder))) {
		if (speed_force || t - decoder->speed_t <= DCC_SPEED_T) {
			switch(decoder->speed_type) {
				case DCC_SPEED_TYPE_NONE:
					/* Nothing, no speed to set */
					break;
				case DCC_SPEED_TYPE_14:
					printf("14 speed not supported yet\n");
					break;
				case DCC_SPEED_TYPE_28:
					dcc_speed28_send(fd, &decoder->address, decoder->speed);
					break;
				case DCC_SPEED_TYPE_128:
					dcc_speed128_send(fd, &decoder->address, decoder->speed);
					break;
			}
			update_done = true;
		}
		if (func_force || t - decoder->func_t <= DCC_FUNCTION_T) {
			dcc_functions_send(fd, decoder, func_force);
			update_done = true;
		}
		if (decoder->func_changed &&
		    t - decoder->func_t > DCC_FUNCTION_T) {
			decoder->func_changed = 0;
		}
	}

	while ((acc = dcc_accessory_decoder_loop(acc))) {
		if (accessory_force || t - acc->value_t <= DCC_ACCESSORY_T) {
			dcc_accessory_decoder_send(fd, acc);
			update_done = true;
		}
	}
	
	return update_done;
}


void *dcc_update_thread(void *arg)
{
	struct timespec ts;
	bool update_done;
	bool func_force;
	bool speed_force;
	bool accessory_force;
	time_t prevt, now;
	int updates = 0;
	int nops = 0;
	
	ts.tv_sec = 0;
	ts.tv_nsec = 100*1000 *1000;
	prevt = time(NULL);
	while (1) {
		update_done = false;
		speed_force = (updates % DCC_SPEED_INTERVAL) == 0;
		func_force = (updates % DCC_FUNCTION_INTERVAL) == 0;
		accessory_force = (updates % DCC_ACCESSORY_INTERVAL) == 0;
		if (rail_power != cur_power) {
			dcc_power(fd_dcc, rail_power);
			cur_power = rail_power;
		} else if (rail_power) {
			update_done = dcc_update(fd_dcc, 
			    speed_force, func_force, accessory_force);
		}
		if (!update_done) {
			nops++;
			if (nops > DCC_SPEED_INTERVAL) {
				nanosleep(&ts, NULL);
				nops = 0;
			}
		}
		dcc_ack_flush(fd_dcc);
		
		updates++;
		
		now = time(NULL);
		if (now != prevt) {
			prevt = now;
//			printf("%d dcc updates per second\n", updates);
			updates = 0;
		}
	}
	return NULL;
}

int main(int argc, char **argv)
{
	struct termios oldtio,newtio;
	struct dcc_decoder *decoder = NULL;
	pthread_t thread;
	pthread_attr_t attr;
	double speed = 0.0, newspeed;
	
	if (argc < 2) {
		printf("No serial device given\n");
		exit(1);
	}
	
	fd_dcc = open(argv[1], O_RDWR | O_NOCTTY ); 
	if (fd_dcc <0) {perror(argv[1]); exit(-1); }

	tcgetattr(fd_dcc, &oldtio); /* save current serial port settings */
	memset(&newtio, 0, sizeof(newtio)); /* clear struct for new port settings */

	newtio.c_cflag = BAUDRATE | CS8 | CLOCAL | CREAD;
	newtio.c_iflag = IGNPAR | ICRNL;
	newtio.c_oflag = 0;
 
	tcflush(fd_dcc, TCIFLUSH);
	tcsetattr(fd_dcc, TCSANOW, &newtio);

	tcgetattr(0, &oldtio);
	newtio = oldtio;
	newtio.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP
                | INLCR | IGNCR | ICRNL | IXON);
//	newtio.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
	newtio.c_lflag &= ~(ECHO | ICANON | ISIG | IEXTEN);
	newtio.c_cflag &= ~(CSIZE | PARENB);
	newtio.c_cflag |= CS8;
	tcsetattr(0, TCSANOW, &newtio);
	
	pthread_attr_init(&attr);
	pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN + 65536);
	
	pthread_create(&thread, &attr, dcc_update_thread, NULL);

	pthread_create(&thread, &attr, dcc_srcpd, NULL);
	
	while (1) {
		int in = fgetc(stdin);
		switch (in) {
			case '.':
				if (!decoder) {
					printf("No decoder selected\n");
					break;
				}
				newspeed = speed + SPEEDSTEP;
				if (fabs(newspeed) < SPEEDSTEP)
					newspeed = copysign(0.0, speed);
				if (newspeed > 1.0)
					newspeed = 1.0;
				speed = newspeed;
				printf("Speed set to %f\n", speed);
				dcc_decoder_speed_set(decoder, speed);
				break;
			case ',':
				if (!decoder) {
					printf("No decoder selected\n");
					break;
				}
				newspeed = speed - SPEEDSTEP;
				if (fabs(newspeed) < SPEEDSTEP)
					newspeed = copysign(0.0, speed);
				if (newspeed < -1.0)
					newspeed = -1.0;
				speed = newspeed;
				printf("Speed set to %f\n", speed);
				dcc_decoder_speed_set(decoder, speed);
				break;
			case 'h':
				if (!decoder) {
					printf("No decoder selected\n");
					break;
				}
				if (signbit(speed))
					speed = -0.0;
				else
					speed = 0.0;
				printf("Speed set to zero\n");
				dcc_decoder_speed_set(decoder, speed);
				break;
			case '0':
				if (!decoder) {
					printf("No decoder selected\n");
					break;
				}
				dcc_decoder_function_toggle(decoder, 
				    DCC_FUNCTION_FL);
				printf("Toggle function F0\n");
				break;
			case '1':
				if (!decoder) {
					printf("No decoder selected\n");
					break;
				}
				dcc_decoder_function_toggle(decoder, 
				    DCC_FUNCTION_F1);
				printf("Toggle function F1\n");
				break;
			case '2':
				if (!decoder) {
					printf("No decoder selected\n");
					break;
				}
				dcc_decoder_function_toggle(decoder, 
				    DCC_FUNCTION_F2);
				printf("Toggle function F2\n");
				break;
			case '3':
				if (!decoder) {
					printf("No decoder selected\n");
					break;
				}
				dcc_decoder_function_toggle(decoder, 
				    DCC_FUNCTION_F3);
				printf("Toggle function F3\n");
				break;
			case '4':
				if (!decoder) {
					printf("No decoder selected\n");
					break;
				}
				dcc_decoder_function_toggle(decoder, 
				    DCC_FUNCTION_F4);
				printf("Toggle function F4\n");
				break;
			case '5':
				if (!decoder) {
					printf("No decoder selected\n");
					break;
				}
				dcc_decoder_function_toggle(decoder, 
				    DCC_FUNCTION_F5);
				printf("Toggle function F5\n");
				break;
			case '6':
				if (!decoder) {
					printf("No decoder selected\n");
					break;
				}
				dcc_decoder_function_toggle(decoder, 
				    DCC_FUNCTION_F6);
				printf("Toggle function F6\n");
				break;
			case '7':
				if (!decoder) {
					printf("No decoder selected\n");
					break;
				}
				dcc_decoder_function_toggle(decoder, 
				    DCC_FUNCTION_F7);
				printf("Toggle function F7\n");
				break;
			case '8':
				if (!decoder) {
					printf("No decoder selected\n");
					break;
				}
				dcc_decoder_function_toggle(decoder, 
				    DCC_FUNCTION_F8);
				printf("Toggle function F8\n");
				break;
			case 'c': {
				char buffer[80];
				int cv;
				int val;
				
				if (!decoder) {
					printf("No decoder selected\n");
					break;
				}
				tcsetattr(0, TCSANOW, &oldtio);
				printf("CV nr\n");
				fgets(buffer, 80, stdin);
				cv = atoi(buffer);
				printf("Value for CV%d:\n", cv);
				fgets(buffer, 80, stdin);
				val = atoi(buffer);
				printf("CV%d: %d (0x%02x)\n", cv, val, val);
				dcc_cv_set( 
				     &decoder->address, 
				     cv, 
				     val);
				tcsetattr(0, TCSANOW, &newtio);
			}
				break;
			case 'r': {
				char buffer[80];
				struct dcc_decoder *entry;
				int add;
				tcsetattr(0, TCSANOW, &oldtio);
				printf("Remove, address:\n");
				fgets(buffer, 80, stdin);
				add = atoi(buffer);
				printf("Removing %d\n", add);
				entry = dcc_decoder_find(add);
				dcc_decoder_del(entry);
				dcc_decoder_put(entry);
				if (entry == decoder) {
					decoder = NULL;
				}
				tcsetattr(0, TCSANOW, &newtio);
			}
				break;
			case 'a': {
				char buffer[80];
				int add;
				tcsetattr(0, TCSANOW, &oldtio);
				printf("Add, address:\n");
				fgets(buffer, 80, stdin);
				add = atoi(buffer);
				printf("Adding %d\n", add);
				decoder = dcc_decoder_add(
					add > 127 ? DCC_ADDRESS_LONG :
					DCC_ADDRESS_SHORT, 
					add, 
					DCC_SPEED_TYPE_128,
					13);
				tcsetattr(0, TCSANOW, &newtio);
				break;
			}
			case 's': {
				char buffer[80];
				int add;
				tcsetattr(0, TCSANOW, &oldtio);
				printf("Select, address:\n");
				fgets(buffer, 80, stdin);
				add = atoi(buffer);
				printf("Selecting %d\n", add);
				if (decoder) {
					dcc_decoder_put(decoder);
				}
				decoder = dcc_decoder_find(add);
				if (decoder) {
					printf("Found decoder\n");
				}
				tcsetattr(0, TCSANOW, &newtio);
				break;
			}
			case 'p': {
				rail_power = !rail_power;
				printf("Rail power set to: %s\n",
				    rail_power ? "on" : "off");
				break;
			}
			case 'q':
				tcsetattr(0, TCSANOW, &oldtio);
				exit(0);
			default:
				break;
		}
	}
	return 0;
}

