/*
	dcc-cmd
	
	A simple DCC command station using the dcc-kiss interface.
	
	Copyright Jeroen Vreeken (jeroen@vreeken.net), 2010, 2011, 2012, 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#include "dcc_cmd.h"

struct dcc_decoder *dcc_decoder_list = NULL;
pthread_mutex_t dcc_decoder_list_mutex = PTHREAD_MUTEX_INITIALIZER;

struct dcc_decoder *dcc_decoder_add(enum dcc_address_type type, int address, 
    enum dcc_speed_type speed_type,
    int nr_functions)
{
	struct dcc_decoder *decoder;
	
	decoder = malloc(sizeof(struct dcc_decoder));
	if (!decoder) {
		perror("Could not alloc decoder structure\n");
		return NULL;
	}
	
	decoder->address.type = type;
	decoder->address.address = address;
	if (type == DCC_ADDRESS_LONG) {
		printf("adding decoder with address %d (%d %d)\n",
		    decoder->address.address,
		    (decoder->address.address >> 8) | 0xc0,
		    decoder->address.address & 0xff);
	} else {
		printf("adding decoder with address %d\n",
		    decoder->address.address);
	}
	decoder->speed = 0;
	decoder->speed_type = speed_type;
	decoder->nr_functions = nr_functions;
	decoder->functions = DCC_FUNCTION_FL; /* f0 (lights) */
	decoder->functions |= DCC_FUNCTION_F3; /* f3 (limit speed) */
	decoder->func_t = 0;
	decoder->func_changed = 0;
	printf("%d functions\n", nr_functions);
	
	decoder->lock = 0;
	
	pthread_mutex_lock(&dcc_decoder_list_mutex);
	decoder->refcount = 2;
	decoder->next = dcc_decoder_list;
	dcc_decoder_list = decoder;
	pthread_mutex_unlock(&dcc_decoder_list_mutex);
	
	return decoder;
}

struct dcc_decoder *dcc_decoder_find(int address)
{
	struct dcc_decoder *decoder = NULL;
	
	pthread_mutex_lock(&dcc_decoder_list_mutex);
	for (decoder = dcc_decoder_list; decoder; decoder = decoder->next) {
		if (decoder->address.address == address) {
			decoder->refcount++;
			break;
		}
	}
	pthread_mutex_unlock(&dcc_decoder_list_mutex);
	return decoder;
}

static void dcc_decoder_unlist(struct dcc_decoder *decoder)
{
	struct dcc_decoder **listp;
	
	for (listp = &dcc_decoder_list; *listp; listp = &(*listp)->next) {
		if (*listp == decoder) {
			*listp = decoder->next;
			break;
		}
	}
	decoder->next = NULL;
}

void dcc_decoder_put(struct dcc_decoder *decoder)
{
	pthread_mutex_lock(&dcc_decoder_list_mutex);
	decoder->refcount--;
	if (decoder->refcount == 0) {
		dcc_decoder_unlist(decoder);
		free(decoder);
	}
	pthread_mutex_unlock(&dcc_decoder_list_mutex);
}

void dcc_decoder_del(struct dcc_decoder *decoder)
{
	pthread_mutex_lock(&dcc_decoder_list_mutex);
	dcc_decoder_unlist(decoder);
	pthread_mutex_unlock(&dcc_decoder_list_mutex);
	dcc_decoder_put(decoder);
}

struct dcc_decoder *dcc_decoder_loop(struct dcc_decoder *decoder)
{
	struct dcc_decoder *next = NULL;

	if (!decoder) {
		pthread_mutex_lock(&dcc_decoder_list_mutex);
		next = dcc_decoder_list;
		if (next) {
			next->refcount++;
		}
		pthread_mutex_unlock(&dcc_decoder_list_mutex);
	} else {
		pthread_mutex_lock(&dcc_decoder_list_mutex);
		next = decoder->next;
		if (next) {
			next->refcount++;
		}
		pthread_mutex_unlock(&dcc_decoder_list_mutex);
		
		dcc_decoder_put(decoder);
	}

	return next;
}

void dcc_decoder_speed_set(struct dcc_decoder *decoder, double speed)
{
	if (speed > 1.0)
		speed = 1.0;
	if (speed < -1.0)
		speed = -1.0;

	switch (decoder->speed_type) {
		case DCC_SPEED_TYPE_14:
		case DCC_SPEED_TYPE_28:
		case DCC_SPEED_TYPE_128:
			decoder->speed = speed;
			decoder->speed_t = time(NULL);
			break;
		default:
			printf("Setting speed not supported!\n");
			break;
	}
//	printf("Decoder %d (@%p) speed: %g\n", decoder->address.address, decoder, speed);
}
