/*
	dcc-cmd
	
	A simple DCC command station using the dcc-kiss interface.
	
	Copyright Jeroen Vreeken (jeroen@vreeken.net), 2010, 2011, 2012, 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>

#include "dcc_interface.h"

#define SLIP_END 192
#define SLIP_ESC 219
#define SLIP_ESC_END 220
#define SLIP_ESC_ESC 221

pthread_mutex_t dcc_fd_mutex = PTHREAD_MUTEX_INITIALIZER;


int slip_add(unsigned char *buf, int pos, unsigned char c)
{
	switch(c) {
		case SLIP_END:
			buf[pos] = SLIP_ESC;
			pos++;
			buf[pos] = SLIP_ESC_END;
			pos++;
			break;
		case SLIP_ESC:
			buf[pos] = SLIP_ESC;
			pos++;
			buf[pos] = SLIP_ESC_ESC;
			pos++;
			break;
		default:
			buf[pos] = c;
			pos++;
			break;
	}
	return pos;
}

#define ACK_PACKET	0
#define ACK_POWER_ON	1
#define ACK_POWER_OFF	2
#define ACK_CUTOUT	3
#define NACK	0xff

#define PACKET_HEADER_LEN	1
#define PACKET_HEADER_DCC	0
#define PACKET_HEADER_POWER_ON	1
#define PACKET_HEADER_POWER_OFF	2
#define PACKET_HEADER_CUTOUT	3

static bool dcc_ack_pending = 0;
static size_t dcc_ack_pending_size = 0;
static struct timespec dcc_ack_pending_tm;

int dcc_recv_ack(int fd, unsigned char *ack, size_t length, struct timespec *t)
{
	size_t len = 0;
	
	while (len < length) {
		unsigned char c;
		int r;
		fd_set set_r;
		
		FD_ZERO(&set_r);
		FD_SET(fd, &set_r);
		
		r = pselect(fd+1, &set_r, NULL, NULL, t, NULL);
		
		/* timeout */
		if (r == 0)
			return 0;
		
		r = read(fd, &c, 1);
		if (r == 1) {
//			printf("c: 0x%02x\n", c);
			switch(c) {
				case SLIP_END:
					if (len) {
						length = len;;
					}
					break;
				default:
					*ack = c;
					ack++;
					len++;
					break;
			}
		} else if (r <= 0 && errno != EAGAIN) {
			return -1;
		}
	}
	
	return len;
}



int dcc_ignore_ack(int fd, size_t length, struct timespec *t)
{
	if (dcc_ack_pending) {
		unsigned char buf[dcc_ack_pending_size];
		dcc_recv_ack(fd, buf, dcc_ack_pending_size, &dcc_ack_pending_tm);
	}
	dcc_ack_pending = true;
	dcc_ack_pending_size = length;
	dcc_ack_pending_tm = *t;
	
	return 0;
}


int dcc_wait_ack(int fd, unsigned char *ack, size_t length, struct timespec *t)
{
	if (dcc_ack_pending) {
		unsigned char buf[dcc_ack_pending_size];
		dcc_recv_ack(fd, buf, dcc_ack_pending_size, &dcc_ack_pending_tm);
		
		dcc_ack_pending = false;
	}
	
	
	return dcc_recv_ack(fd, ack, length, t);
}

int dcc_ack_flush(int fd)
{
	unsigned char ack;
	int r;

	if (dcc_ack_pending) {
		unsigned char buf[dcc_ack_pending_size];
		dcc_recv_ack(fd, buf, dcc_ack_pending_size, &dcc_ack_pending_tm);
		
		dcc_ack_pending = false;
	}
	do {
		r = dcc_wait_ack(fd, &ack, 1, &(struct timespec){ 0, 0});
		if (r == 1) {
			printf("Spurious Ack: %d\n", ack);
		}
	} while (r == 1);

	return 0;
}

int dcc_power(int fd, bool power)
{
	unsigned char packet_bytes[4];
	unsigned char ack;
	int len = 0, r;


	packet_bytes[len] = SLIP_END;
	len++;
	len = slip_add(packet_bytes, len, 
	    power ? PACKET_HEADER_POWER_ON : PACKET_HEADER_POWER_OFF);
	packet_bytes[len] = SLIP_END;
	len++;

	pthread_mutex_lock(&dcc_fd_mutex);

	write(fd, packet_bytes, len);

	r = dcc_wait_ack(fd, &ack, 1, &(struct timespec){ 0, 100*1000*1000});
	if (r != 1) {
		printf("Ack timeout\n");
	}
	if (ack != (power ? ACK_POWER_ON : ACK_POWER_OFF)) {
		printf("Unexpected ack value: 0x%02x\n", ack);
	}

	pthread_mutex_unlock(&dcc_fd_mutex);
	return 0;
}

bool debug_send = false;

int dcc_send_real(int fd, unsigned char *packet, size_t length, struct dcc_address *address, bool wait_ack)
{
	unsigned char packet_bytes[20];
	int i, len = 0, r;
	unsigned char checksum = 0;
	unsigned char ack;

	checksum = 0;
	
	packet_bytes[len] = SLIP_END;
	len++;
	/* packet header */
	len = slip_add(packet_bytes, len, PACKET_HEADER_DCC);
	if (address->type == DCC_ADDRESS_SHORT) {
		len = slip_add(packet_bytes, len, address->address);
		checksum ^= address->address;
	} else if (address->type == DCC_ADDRESS_LONG) {
		len = slip_add(packet_bytes, len, 0xc0 | (address->address >> 8));
		checksum ^= 0xc0 | (address->address >> 8);
		len = slip_add(packet_bytes, len, address->address & 0xff);
		checksum ^= (address->address & 0xff);
	} else if (address->type == DCC_ADDRESS_ACCESSORY) {
		unsigned char add1;
		unsigned char add2;
		unsigned char add3;
		
		add1 = ((address->address + 3) / 4) & 0x3f;
		add2 = ~((address->address + 3) / 256) & 0x7;
		add3 = (address->address + 3) & 0x3;
		
		len = slip_add(packet_bytes, len, 0x80 | add1);
		checksum ^= 0x80 | add1;

		packet[0] &= 0x89;
		packet[0] |= (add2 < 4) | (add3 << 1);
	} else {
		printf("Unsupported address type: %d\n", address->type);
		return -1;
	}
	
	for (i = 0; i < length; i++) {
		len = slip_add(packet_bytes, len, packet[i]);
		checksum ^= packet[i];
		if (debug_send)
			printf("%02x ", packet[i]);
	}
	len = slip_add(packet_bytes, len, checksum);
	packet_bytes[len] = SLIP_END;
	len++;

	write(fd, packet_bytes, len);
	if (debug_send)
		printf("Packet send\n");

	if (!wait_ack) {
		r = dcc_ignore_ack(fd, 1, &(struct timespec){ 0, 100*1000*1000});
	} else {
		r = dcc_wait_ack(fd, &ack, 1, &(struct timespec){ 0, 100*1000*1000});
		if (r != 1) {
			printf("Ack timeout\n");
		}
		if (ack != ACK_PACKET) {
			printf("Unexpected ack value: 0x%02x\n", ack);
		}
		if (debug_send)
			printf("Ack done\n");
	}

	return length;
}

int dcc_send_lock(void)
{
	pthread_mutex_lock(&dcc_fd_mutex);

	return 0;
}

int dcc_send_unlock(void)
{
	pthread_mutex_unlock(&dcc_fd_mutex);

	return 0;
}

int dcc_send(int fd, unsigned char *packet, size_t length, struct dcc_address *address, bool wait_ack)
{
	int ret;
	
	dcc_send_lock();

	ret = dcc_send_real(fd, packet, length, address, wait_ack);
	
	dcc_send_unlock();
	
	return ret;
}
