/*
	dcc_srcpd.c
		
	SRCP implementation for a simple DCC command station

	Copyright Jeroen Vreeken (jeroen@vreeken.net), 2010, 2012, 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <netinet/in.h>
#include <pthread.h>
#include <string.h>
#include <stdbool.h>
#include <unistd.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <math.h>

#include "dcc_srcpd.h"
#include "dcc_srcpd_info_q.h"
#include "dcc_srcpd_gl.h"
#include "dcc_srcpd_ga.h"
#include "dcc_srcpd_power.h"
#include "dcc_srcpd_sm.h"
#include "dcc_srcpd_description.h"
#include "dcc_srcpd_session.h"
#include "dcc_srcpd_gm.h"
#include "dcc_srcpd_lock.h"
#include "dcc_cmd.h"

enum dcc_srcpd_mode {
	DCC_SRCPD_MODE_COMMAND,
	DCC_SRCPD_MODE_INFO,
};

struct dcc_srcpd_client {
	bool go;
	enum dcc_srcpd_mode mode;
	int session_id;
};

static unsigned int dcc_srcpd_session_id = 1;

static pthread_mutex_t dcc_srcpd_session_id_mutex = PTHREAD_MUTEX_INITIALIZER;

void dcc_srcpd_rep_timestamp(FILE *fd_sock)
{
	struct timeval tv;
	
	gettimeofday(&tv, NULL);
	
	fprintf(fd_sock, "%ld.%06ld ", (long)tv.tv_sec, (long)tv.tv_usec);
}

char *dcc_srcpd_rep_timestamp_str(char *buffer)
{
	struct timeval tv;
	
	gettimeofday(&tv, NULL);
	
	buffer += sprintf(buffer, "%ld.%06ld ", (long)tv.tv_sec, (long)tv.tv_usec);
	return buffer;
}



void dcc_srcpd_ok(FILE *fd_sock)
{
	dcc_srcpd_rep_timestamp(fd_sock);
	fprintf(fd_sock, "%s\n", DCC_SRCPD_REP_OK);
}


void dcc_srcpd_no_data(FILE *fd_sock)
{
	dcc_srcpd_rep_timestamp(fd_sock);
	fprintf(fd_sock, "%s\n", DCC_SRCPD_REP_NO_DATA);
}

void dcc_srcpd_wrong_value(FILE *fd_sock)
{
	dcc_srcpd_rep_timestamp(fd_sock);
	fprintf(fd_sock, "%s\n", DCC_SRCPD_REP_WRONG_VALUE);
}

void dcc_srcpd_unsupported_device(FILE *fd_sock)
{
	dcc_srcpd_rep_timestamp(fd_sock);
	fprintf(fd_sock, "%s\n", DCC_SRCPD_REP_UNSUPPORTED_DEVICE);
}

void dcc_srcpd_unsupported_device_protocol(FILE *fd_sock)
{
	dcc_srcpd_rep_timestamp(fd_sock);
	fprintf(fd_sock, "%s\n", DCC_SRCPD_REP_UNSUPPORTED_DEVICE_PROTOCOL);
}

void dcc_srcpd_unsupported(FILE *fd_sock)
{
	dcc_srcpd_rep_timestamp(fd_sock);
	fprintf(fd_sock, "%s\n", DCC_SRCPD_REP_UNSUPPORTED);
}

void dcc_srcpd_device_locked(FILE *fd_sock)
{
	dcc_srcpd_rep_timestamp(fd_sock);
	fprintf(fd_sock, "%s\n", DCC_SRCPD_REP_DEVICE_LOCKED);
}


void dcc_srcpd_info(struct dcc_srcpd_client *client, FILE *fd_sock)
{
	struct dcc_decoder *entry;
	struct dcc_accessory_decoder *aentry;
	int r = 0;
	struct dcc_srcpd_info *info = NULL;
	
	dcc_srcpd_info_q_client_add();

	dcc_srcpd_info_get_description(fd_sock, 0);
	dcc_srcpd_info_get_description(fd_sock, 1);

	dcc_srcpd_info_init_power(fd_sock);
	dcc_srcpd_info_get_power(fd_sock);

	pthread_mutex_lock(&dcc_decoder_list_mutex);
	/* loop over decoderlist */
	for (entry = dcc_decoder_list; entry; entry = entry->next) {
		dcc_srcpd_info_init_gl(fd_sock, entry);
	}
	for (entry = dcc_decoder_list; entry; entry = entry->next) {
		dcc_srcpd_info_get_gl(fd_sock, entry);
		dcc_srcpd_lock_get_gl(fd_sock, entry);
	}
	pthread_mutex_unlock(&dcc_decoder_list_mutex);

	/* loop over decoderlist */
	pthread_mutex_lock(&dcc_accessory_decoder_list_mutex);
	for (aentry = dcc_accessory_decoder_list; aentry; aentry = aentry->next) {
		dcc_srcpd_info_init_ga(fd_sock, aentry);
	}
	for (aentry = dcc_accessory_decoder_list; aentry; aentry = aentry->next) {
		dcc_srcpd_info_get_ga(fd_sock, aentry, 0);
		dcc_srcpd_info_get_ga(fd_sock, aentry, 1);
		dcc_srcpd_lock_get_ga(fd_sock, aentry);
	}
	pthread_mutex_unlock(&dcc_accessory_decoder_list_mutex);
	
	fflush(fd_sock);

	while (1) {
		info = dcc_srcpd_info_q_get(info);

		if (!info->dst_session_id || info->dst_session_id == client->session_id) {
			r = fprintf(fd_sock, "%s", info->buffer);
		
			if (r < 0)
				break;
			r = fflush(fd_sock);
			if (r != 0)
				break;
		}
	}
	dcc_srcpd_info_q_put(info);

	dcc_srcpd_info_q_client_remove();
}

static char *dcc_srcpd_req_token(char *req, int nr)
{
	bool nonws = true;
	
	if (!nr) {
		return req;
	}
	
	while (*req) {
		if (*req != ' ' && *req != '\t') {
			if (!nonws) {
				nr--;
				if (!nr)
					break;
			}
			nonws = true;
		} else {
			nonws = false;
		}
		req++;
	}

	return req;
}

static bool dcc_srcpd_req_token_cmp(char *req, int nr, char *str)
{
	int i;
	bool same=true;
	
	req = dcc_srcpd_req_token(req, nr);
	
	for (i = 0; i < strlen(str); i++) {
		if (req[i] != str[i])
			same = false;
	}
	if (req[i] != ' ' && req[i] != '\t' && req[i] != '\n' && req[i] != '\r')
		same = false;

	return same;
}

static int dcc_srcpd_req_token_int(char *req, int nr)
{
	int i;
	
	req = dcc_srcpd_req_token(req, nr);
	sscanf(req, "%d", &i);
	
	return i;
}


static bool dcc_srcpd_req_handle_set_gl(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	int bus, address, drivemode, V, V_max;
	int f0, f1, f2, f3, f4, f5, f6, f7, f8;
	int f9, f10, f11, f12;
	int f13, f14, f15, f16, f17, f18, f19, f20;
	int f21, f22, f23, f24, f25, f26, f27, f28;
	struct dcc_decoder *decoder;
  	  
	sscanf(buffer, "SET %d GL %d"
	    " %d %d %d"
	    " %d %d %d %d %d"
	    " %d %d %d %d"
	    " %d %d %d %d"
	    " %d %d %d %d %d %d %d %d"
	    " %d %d %d %d %d %d %d %d",
	    &bus,
	    &address,
	    &drivemode, &V, &V_max,
	    &f0, &f1, &f2, &f3, &f4,
	    &f5, &f6, &f7, &f8,
	    &f9, &f10, &f11, &f12, 
	    &f13, &f14, &f15, &f16, &f17, &f18, &f19, &f20, 
	    &f21, &f22, &f23, &f24, &f25, &f26, &f27, &f28 );
	decoder = dcc_decoder_find(address);
	if (!decoder) {
		dcc_srcpd_no_data(fd_sock);
	} else {
		if (decoder->lock && decoder->lock != client->session_id)
			goto err_locked;

		dcc_decoder_speed_set(decoder, (double)V/(double)V_max *
		    (drivemode == 1 ? 1.0 : -1.0));
		dcc_decoder_function_set(decoder, 0, f0);
		dcc_decoder_function_set(decoder, 1, f1);
		dcc_decoder_function_set(decoder, 2, f2);
		dcc_decoder_function_set(decoder, 3, f3);
		dcc_decoder_function_set(decoder, 4, f4);
		dcc_decoder_function_set(decoder, 5, f5);
		dcc_decoder_function_set(decoder, 6, f6);
		dcc_decoder_function_set(decoder, 7, f7);
		dcc_decoder_function_set(decoder, 8, f8);
		dcc_decoder_function_set(decoder, 9, f9);
		dcc_decoder_function_set(decoder, 10, f10);
		dcc_decoder_function_set(decoder, 11, f11);
		dcc_decoder_function_set(decoder, 12, f12);
		dcc_decoder_function_set(decoder, 13, f13);
		dcc_decoder_function_set(decoder, 14, f14);
		dcc_decoder_function_set(decoder, 15, f15);
		dcc_decoder_function_set(decoder, 16, f16);
		dcc_decoder_function_set(decoder, 17, f17);
		dcc_decoder_function_set(decoder, 18, f18);
		dcc_decoder_function_set(decoder, 19, f19);
		dcc_decoder_function_set(decoder, 20, f20);
		dcc_decoder_function_set(decoder, 21, f21);
		dcc_decoder_function_set(decoder, 22, f22);
		dcc_decoder_function_set(decoder, 23, f23);
		dcc_decoder_function_set(decoder, 24, f24);
		dcc_decoder_function_set(decoder, 25, f25);
		dcc_decoder_function_set(decoder, 26, f26);
		dcc_decoder_function_set(decoder, 27, f27);
		dcc_decoder_function_set(decoder, 28, f28);
		dcc_srcpd_info_get_gl_update(decoder);
		dcc_decoder_put(decoder);
  		dcc_srcpd_ok(fd_sock);
  	}

	return true;

err_locked:
	dcc_srcpd_device_locked(fd_sock);
	return true;
}

static bool dcc_srcpd_req_handle_set_ga(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	int bus, address, port, value, delay;
	struct dcc_accessory_decoder *decoder;
  	  
	sscanf(buffer, "SET %d GA %d %d %d %d",
	    &bus,
	    &address,
	    &port,
	    &value,
	    &delay
	);
	decoder = dcc_accessory_decoder_find(address);
	if (!decoder) {
		dcc_srcpd_no_data(fd_sock);
	} else {
		if (decoder->lock && decoder->lock != client->session_id)
			goto err_locked;

		dcc_accessory_decoder_output_set(decoder, port, value);
		dcc_srcpd_info_get_ga_update(decoder, port);
		if (value && delay > 0) {
			dcc_srcpd_ga_delay(decoder, port, delay);
		} else {
			dcc_accessory_decoder_put(decoder);
		}
  		dcc_srcpd_ok(fd_sock);
  	}

	return true;

err_locked:
	dcc_srcpd_device_locked(fd_sock);
	return true;
}

static bool dcc_srcpd_req_handle_get_gl(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	int bus, address;
	struct dcc_decoder *decoder;
	
	sscanf(buffer, "GET %d GL %d",
	    &bus,
	    &address);
	decoder = dcc_decoder_find(address);
	if (!decoder) {
		dcc_srcpd_no_data(fd_sock);
	} else {
		dcc_srcpd_info_get_gl(fd_sock, decoder);
		dcc_decoder_put(decoder);
	}

	return true;
}

static bool dcc_srcpd_req_handle_get_ga(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	int bus, address, port;
	struct dcc_accessory_decoder *decoder;
	
	sscanf(buffer, "GET %d GA %d %d",
	    &bus,
	    &address,
	    &port);
	decoder = dcc_accessory_decoder_find(address);
	if (!decoder) {
		dcc_srcpd_no_data(fd_sock);
	} else {
		dcc_srcpd_info_get_ga(fd_sock, decoder, port);
		dcc_accessory_decoder_put(decoder);
	}

	return true;
}

static bool dcc_srcpd_req_handle_set_gm(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	int bus, send_to, reply_to;
  	char msg[strlen(buffer)+1];
	
	printf("%s", buffer);
	
	sscanf(buffer, "SET %d GM %d %d %[^\n]",
	    &bus,
	    &send_to,
	    &reply_to,
	    msg
	);

	dcc_srcpd_info_get_gm_update(send_to, reply_to, msg);
	dcc_srcpd_ok(fd_sock);

	return true;
}

static bool dcc_srcpd_req_handle_term_gl(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	int bus, address;
	struct dcc_decoder *decoder;
		
	sscanf(buffer, "TERM %d GL %d",
	    &bus, &address);
	decoder = dcc_decoder_find(address);
	if (!decoder) {
		dcc_srcpd_no_data(fd_sock);
	} else {
		dcc_srcpd_info_term_gl_update(decoder);
		dcc_decoder_del(decoder);
		dcc_decoder_put(decoder);
		dcc_srcpd_ok(fd_sock);
	}

	return true;
}

static bool dcc_srcpd_req_handle_init_gl(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	int bus, address, protocolversion, speedsteps;
	int nroffunctions;
	struct dcc_decoder *decoder;
	
	sscanf(buffer, "INIT %d GL %d N %d %d %d",
	    &bus, 
	    &address, 
	    &protocolversion, 
	    &speedsteps,
	    &nroffunctions);
	decoder = dcc_decoder_find(address);
	if (!decoder) {
		decoder = dcc_decoder_add(
		    protocolversion == 1 ? DCC_ADDRESS_SHORT : DCC_ADDRESS_LONG,
		    address,
		    speedsteps >= 126 ? DCC_SPEED_TYPE_128 :
		    speedsteps == 28 ? DCC_SPEED_TYPE_28 :
		    speedsteps == 14 ? DCC_SPEED_TYPE_14 :
		    DCC_SPEED_TYPE_NONE,
		    nroffunctions);

		dcc_srcpd_info_init_gl_update(decoder);
		dcc_srcpd_info_get_gl_update(decoder);
	}
	dcc_decoder_put(decoder);
	dcc_srcpd_ok(fd_sock);

	return true;
}

static bool dcc_srcpd_req_handle_init_ga(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	int bus, address;
	struct dcc_accessory_decoder *decoder;
	
	sscanf(buffer, "INIT %d GA %d N",
	    &bus, 
	    &address);
	decoder = dcc_accessory_decoder_find(address);
	if (!decoder) {
		decoder = dcc_accessory_decoder_add(DCC_ADDRESS_ACCESSORY, address);

		dcc_srcpd_info_init_ga_update(decoder);
		dcc_srcpd_info_get_ga_update(decoder, 0);
		dcc_srcpd_info_get_ga_update(decoder, 1);
	}
	dcc_accessory_decoder_put(decoder);
	dcc_srcpd_ok(fd_sock);

	return true;
}

static bool dcc_srcpd_req_handle_term_ga(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	int bus, address;
	struct dcc_accessory_decoder *decoder;
		
	sscanf(buffer, "TERM %d GA %d",
	    &bus, &address);
	decoder = dcc_accessory_decoder_find(address);
	if (!decoder) {
		dcc_srcpd_no_data(fd_sock);
	} else {
		dcc_srcpd_info_term_ga_update(decoder);
		dcc_accessory_decoder_del(decoder);
		dcc_accessory_decoder_put(decoder);
		dcc_srcpd_ok(fd_sock);
	}

	return true;
}

static bool dcc_srcpd_req_handle_term_power(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	dcc_srcpd_info_term_power_update();
	dcc_srcpd_ok(fd_sock);

	return true;
}

static bool dcc_srcpd_req_handle_init_power(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	dcc_srcpd_info_init_power_update();
	dcc_srcpd_info_get_power_update();
	dcc_srcpd_ok(fd_sock);

	return true;
}

static bool dcc_srcpd_req_handle_set_power(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	if (dcc_srcpd_req_token_cmp(buffer, 3, DCC_SRCPD_ON)) {
		rail_power = true;
	} else if (dcc_srcpd_req_token_cmp(buffer, 3, DCC_SRCPD_OFF)) {
		rail_power = false;
	}
	dcc_srcpd_info_get_power_update();
	dcc_srcpd_ok(fd_sock);

	return true;
}

static bool dcc_srcpd_req_handle_get_power(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	dcc_srcpd_info_get_power(fd_sock);

	return true;
}

static bool dcc_srcpd_req_handle_init_sm(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	if (!dcc_srcpd_req_token_cmp(buffer, 3, DCC_SRCPD_NMRA)) {
		dcc_srcpd_unsupported_device_protocol(fd_sock);
		
		return true;
	}
	
	dcc_srcpd_info_init_sm_update();
	dcc_srcpd_ok(fd_sock);

	return true;
}

static bool dcc_srcpd_req_handle_term_sm(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	dcc_srcpd_info_term_sm_update();
	dcc_srcpd_ok(fd_sock);

	return true;
}

static bool dcc_srcpd_req_handle_set_sm(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	int bus, address, value, cv;
	struct dcc_decoder *decoder;
  	  
	if (!dcc_srcpd_req_token_cmp(buffer, 4, DCC_SRCPD_CV)) {
		dcc_srcpd_unsupported(fd_sock);
		
		return true;
	}

	sscanf(buffer, "SET %d SM %d CV %d %d",
	    &bus,
	    &address,
	    &cv,
	    &value
	);
	
	decoder = dcc_decoder_find(address);
	if (!decoder) {
		dcc_srcpd_unsupported_device(fd_sock);
		
		return true;
	}

	if (decoder->lock && decoder->lock != client->session_id)
		goto err_locked;
	
	dcc_cv_set(&decoder->address, cv, value);

	dcc_srcpd_info_get_sm_update(address, cv, value);
	dcc_decoder_put(decoder);
	dcc_srcpd_ok(fd_sock);

	return true;

err_locked:
	dcc_srcpd_device_locked(fd_sock);
	return true;
}


static bool dcc_srcpd_req_handle_get_description(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	int bus = dcc_srcpd_req_token_int(buffer, 1);

	dcc_srcpd_info_get_description(fd_sock, bus);
	
	return true;
}

static bool dcc_srcpd_req_handle_get_session(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	int session_id = dcc_srcpd_req_token_int(buffer, 3);

	dcc_srcpd_info_get_session(fd_sock, session_id);
	
	return true;
}

static bool dcc_srcpd_req_handle_term_session(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	dcc_srcpd_ok(fd_sock);

	return false;
}

static bool dcc_srcpd_req_handle_set_lock(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	int address = dcc_srcpd_req_token_int(buffer, 4);
	int duration = dcc_srcpd_req_token_int(buffer, 5);
	int ret = -1;
	
	if (duration)
		goto ret_err;

	if (dcc_srcpd_req_token_cmp(buffer, 3, DCC_SRCPD_GROUP_GL)) {
		struct dcc_decoder *decoder;

		decoder = dcc_decoder_find(address);
		if (decoder) {
			ret = dcc_srcpd_lock_set_gl(fd_sock, decoder, duration, client->session_id);
			dcc_decoder_put(decoder);
		}
	} else if (dcc_srcpd_req_token_cmp(buffer, 3, DCC_SRCPD_GROUP_SM)) {
		struct dcc_decoder *decoder;

		decoder = dcc_decoder_find(address);
		if (decoder) {
			ret = dcc_srcpd_lock_set_sm(fd_sock, decoder, duration, client->session_id);
			dcc_decoder_put(decoder);
		}
	} else if (dcc_srcpd_req_token_cmp(buffer, 3, DCC_SRCPD_GROUP_GA)) {
		struct dcc_accessory_decoder *decoder;

		decoder = dcc_accessory_decoder_find(address);
		if (decoder) {
			ret = dcc_srcpd_lock_set_ga(fd_sock, decoder, duration, client->session_id);
			dcc_accessory_decoder_put(decoder);
		}
	}

ret_err:
	if (!ret)
		dcc_srcpd_wrong_value(fd_sock);
	else
		dcc_srcpd_ok(fd_sock);

	return true;
}

static bool dcc_srcpd_req_handle_get_lock(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	int address = dcc_srcpd_req_token_int(buffer, 4);
	int ret = -1;
	
	if (dcc_srcpd_req_token_cmp(buffer, 3, DCC_SRCPD_GROUP_GL)) {
		struct dcc_decoder *decoder;

		decoder = dcc_decoder_find(address);
		if (decoder) {
			ret = dcc_srcpd_lock_get_gl(fd_sock, decoder);
			dcc_decoder_put(decoder);
		}
	} else if (dcc_srcpd_req_token_cmp(buffer, 3, DCC_SRCPD_GROUP_SM)) {
		struct dcc_decoder *decoder;

		decoder = dcc_decoder_find(address);
		if (decoder) {
			ret = dcc_srcpd_lock_get_sm(fd_sock, decoder);
			dcc_decoder_put(decoder);
		}
	}
	if (dcc_srcpd_req_token_cmp(buffer, 3, DCC_SRCPD_GROUP_GA)) {
		struct dcc_accessory_decoder *decoder;

		decoder = dcc_accessory_decoder_find(address);
		if (decoder) {
			ret = dcc_srcpd_lock_get_ga(fd_sock, decoder);
			dcc_accessory_decoder_put(decoder);
		}
	}

	if (!ret)
		dcc_srcpd_wrong_value(fd_sock);

	return true;
}

static bool dcc_srcpd_req_handle_term_lock(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	int address = dcc_srcpd_req_token_int(buffer, 4);
	bool ret = false;
	
	if (dcc_srcpd_req_token_cmp(buffer, 3, DCC_SRCPD_GROUP_GL)) {
		struct dcc_decoder *decoder;

		decoder = dcc_decoder_find(address);
		if (decoder) {
			ret = dcc_srcpd_lock_term_gl(fd_sock, client->session_id, decoder);
			dcc_decoder_put(decoder);
		}
	} else if (dcc_srcpd_req_token_cmp(buffer, 3, DCC_SRCPD_GROUP_SM)) {
		struct dcc_decoder *decoder;

		decoder = dcc_decoder_find(address);
		if (decoder) {
			ret = dcc_srcpd_lock_term_sm(fd_sock, client->session_id, decoder);
			dcc_decoder_put(decoder);
		}
	} else if (dcc_srcpd_req_token_cmp(buffer, 3, DCC_SRCPD_GROUP_GA)) {
		struct dcc_accessory_decoder *decoder;

		decoder = dcc_accessory_decoder_find(address);
		if (decoder) {
			ret = dcc_srcpd_lock_term_ga(fd_sock, client->session_id, decoder);
			dcc_accessory_decoder_put(decoder);
		}
	}

	if (ret)
		dcc_srcpd_ok(fd_sock);
	else
		dcc_srcpd_wrong_value(fd_sock);

	return true;
}

static bool dcc_srcpd_req_handle_check(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	dcc_srcpd_ok(fd_sock);
	return true;
}

static bool dcc_srcpd_req_handle_go(struct dcc_srcpd_client *client, FILE *fd_sock, char *buffer)
{
	unsigned int session_id;

	pthread_mutex_lock(&dcc_srcpd_session_id_mutex);
	if (dcc_srcpd_session_id < 1) {
		dcc_srcpd_session_id = 1;
	}
	session_id = dcc_srcpd_session_id++;
	pthread_mutex_unlock(&dcc_srcpd_session_id_mutex);
	
	client->session_id = session_id;

	dcc_srcpd_rep_timestamp(fd_sock);
	fprintf(fd_sock, "%s %d\n",
	    DCC_SRCPD_REP_GO_OK,
	    session_id);

	dcc_srcpd_info_init_session_update(client->session_id);

	if (client->mode == DCC_SRCPD_MODE_INFO) {
		dcc_srcpd_info(client, fd_sock);
		return false;
	}
	client->go = true;
	return true;
}


struct req_entry {
	bool go_state;
	char *token0;
	char *token2;
	bool (*req_handle)(struct dcc_srcpd_client *, FILE *, char *);
};

struct req_entry req_list[] = {
	{ true, DCC_SRCPD_REQ_SET, DCC_SRCPD_GROUP_GL, dcc_srcpd_req_handle_set_gl },
	{ true, DCC_SRCPD_REQ_SET, DCC_SRCPD_GROUP_GA, dcc_srcpd_req_handle_set_ga },
	{ true, DCC_SRCPD_REQ_SET, DCC_SRCPD_GROUP_POWER, dcc_srcpd_req_handle_set_power },
	{ true, DCC_SRCPD_REQ_SET, DCC_SRCPD_GROUP_GM, dcc_srcpd_req_handle_set_gm },
	{ true, DCC_SRCPD_REQ_SET, DCC_SRCPD_GROUP_SM, dcc_srcpd_req_handle_set_sm },
	{ true, DCC_SRCPD_REQ_SET, DCC_SRCPD_GROUP_LOCK, dcc_srcpd_req_handle_set_lock },

	{ false, DCC_SRCPD_REQ_GO, NULL, dcc_srcpd_req_handle_go },

	{ true, DCC_SRCPD_REQ_INIT, DCC_SRCPD_GROUP_GL, dcc_srcpd_req_handle_init_gl },
	{ true, DCC_SRCPD_REQ_INIT, DCC_SRCPD_GROUP_GA, dcc_srcpd_req_handle_init_ga },
	{ true, DCC_SRCPD_REQ_INIT, DCC_SRCPD_GROUP_POWER, dcc_srcpd_req_handle_init_power },
	{ true, DCC_SRCPD_REQ_INIT, DCC_SRCPD_GROUP_SM, dcc_srcpd_req_handle_init_sm },

	{ true, DCC_SRCPD_REQ_GET, DCC_SRCPD_GROUP_GL, dcc_srcpd_req_handle_get_gl },
	{ true, DCC_SRCPD_REQ_GET, DCC_SRCPD_GROUP_GA, dcc_srcpd_req_handle_get_ga },
	{ true, DCC_SRCPD_REQ_GET, DCC_SRCPD_GROUP_POWER, dcc_srcpd_req_handle_get_power },
	{ true, DCC_SRCPD_REQ_GET, DCC_SRCPD_GROUP_DESCRIPTION, dcc_srcpd_req_handle_get_description },
	{ true, DCC_SRCPD_REQ_GET, DCC_SRCPD_GROUP_SESSION, dcc_srcpd_req_handle_get_session },
	{ true, DCC_SRCPD_REQ_GET, DCC_SRCPD_GROUP_LOCK, dcc_srcpd_req_handle_get_lock },

	{ true, DCC_SRCPD_REQ_TERM, DCC_SRCPD_GROUP_GL, dcc_srcpd_req_handle_term_gl },
	{ true, DCC_SRCPD_REQ_TERM, DCC_SRCPD_GROUP_GA, dcc_srcpd_req_handle_term_ga },
	{ true, DCC_SRCPD_REQ_TERM, DCC_SRCPD_GROUP_POWER, dcc_srcpd_req_handle_term_power },
	{ true, DCC_SRCPD_REQ_TERM, DCC_SRCPD_GROUP_SM, dcc_srcpd_req_handle_term_sm },
	{ true, DCC_SRCPD_REQ_TERM, DCC_SRCPD_GROUP_SESSION, dcc_srcpd_req_handle_term_session },
	{ true, DCC_SRCPD_REQ_TERM, DCC_SRCPD_GROUP_LOCK, dcc_srcpd_req_handle_term_lock },

	{ true, DCC_SRCPD_REQ_CHECK, NULL, dcc_srcpd_req_handle_check },
};

		

void *dcc_srcpd_client(void *arg)
{
	int sock = *(int *)arg;
	char buffer[DCC_SRCPD_REQ_LEN];
	FILE *fd_sock;
	int i;
	struct dcc_srcpd_client client;
	free(arg);
	
	client.go = false;
	client.mode = DCC_SRCPD_MODE_COMMAND;
	
	fd_sock = fdopen(sock, "w+");
	if (!fd_sock) {
		perror("fdopen()");
		return NULL;
	}
	
	fprintf(fd_sock, "%s", DCC_SRCPD_VERSION);
	fflush(fd_sock);
	
	while (fgets(buffer, DCC_SRCPD_REQ_LEN, fd_sock)) {
		bool handled = false;
		
		if (!client.go && 
		    dcc_srcpd_req_token_cmp(buffer, 0, DCC_SRCPD_REQ_SET) &&
		    dcc_srcpd_req_token_cmp(buffer, 1, DCC_SRCPD_PROTOCOL) &&
		    dcc_srcpd_req_token_cmp(buffer, 2, DCC_SRCPD_SRCP)) {
			/* todo: actually check procotol version */
			dcc_srcpd_rep_timestamp(fd_sock);
			fprintf(fd_sock, "%s\n", 
			    DCC_SRCPD_REP_SET_PROTOCOL_OK);
			handled = true;
		}
		if (!client.go && 
		    dcc_srcpd_req_token_cmp(buffer, 0, DCC_SRCPD_REQ_SET) &&
		    dcc_srcpd_req_token_cmp(buffer, 1, DCC_SRCPD_CONNECTIONMODE) &&
		    dcc_srcpd_req_token_cmp(buffer, 2, DCC_SRCPD_SRCP)) {
			if (dcc_srcpd_req_token_cmp(buffer, 3, DCC_SRCPD_INFO)) {
				client.mode = DCC_SRCPD_MODE_INFO;
				dcc_srcpd_rep_timestamp(fd_sock);
				fprintf(fd_sock, "%s\n",
				    DCC_SRCPD_REP_SET_CONNECTIONMODE_OK);
			} else if (dcc_srcpd_req_token_cmp(buffer, 3, DCC_SRCPD_COMMAND)) {
				client.mode = DCC_SRCPD_MODE_COMMAND;
				dcc_srcpd_rep_timestamp(fd_sock);
				fprintf(fd_sock, "%s\n",
				    DCC_SRCPD_REP_SET_CONNECTIONMODE_OK);
			} else {
				dcc_srcpd_rep_timestamp(fd_sock);
				fprintf(fd_sock, "%s\n",
				    DCC_SRCPD_REP_SET_CONNECTIONMODE_ERROR);
			}
			handled = true;
		}

		if (!handled) for (i = 0; i < sizeof(req_list)/sizeof(struct req_entry); i++) {
			if (req_list[i].go_state == client.go &&
			    (!req_list[i].token0 || dcc_srcpd_req_token_cmp(buffer, 0, req_list[i].token0)) &&
			    (!req_list[i].token2 || dcc_srcpd_req_token_cmp(buffer, 2, req_list[i].token2))) {
			    	bool ret;
				
				ret = req_list[i].req_handle(&client, fd_sock, buffer);
				handled = true;
				if (!ret) {
					printf("SRCP client gone\n");
					goto close_session;
				}
				break;
			}
		}
		
		if (!handled) {
			printf("unknown command: %s\n", buffer);
			dcc_srcpd_rep_timestamp(fd_sock);
			fprintf(fd_sock, "%s\n", DCC_SRCPD_REP_UNKNOWN_COMMAND);
		}
		fflush(fd_sock);
	}
	
	printf("SRCP client lost\n");

close_session:
	dcc_srcpd_lock_session_clean(client.session_id);

	fclose(fd_sock);
	dcc_srcpd_info_term_session_update(client.session_id);

	return NULL;
}

void *dcc_srcpd(void *arg)
{
	struct sockaddr_in6 from6;
	struct sockaddr_in6 sin6;
	struct sockaddr_in from;
	struct sockaddr_in sin;
	int listensock, listensock6;
	pthread_t thread;
	pthread_attr_t attr;
	int i;
	int nfds;
	
	signal(SIGPIPE, SIG_IGN);

	memset(&sin6, 0, sizeof(sin6));
	memset(&sin, 0, sizeof(sin));
	sin6.sin6_family = AF_INET6;
	sin6.sin6_port = htons(DCC_SRCPD_PORT);
	sin6.sin6_addr = in6addr_any;
	sin.sin_family = AF_INET;
	sin.sin_port = htons(DCC_SRCPD_PORT);
	sin.sin_addr.s_addr = htonl(INADDR_ANY);
	
	listensock6 = socket(AF_INET6, SOCK_STREAM, 0);

	if (listensock6 < 0) {
		perror("IPv6 socket() failed");
	}

	i = 1;
	setsockopt (listensock6, SOL_SOCKET, SO_REUSEADDR, &i, sizeof (int));

	if (bind(listensock6, (struct sockaddr *)&sin6, sizeof(sin6)) < 0) {
		perror("IPv6 bind() failed");
	}
	listen(listensock6, 10);

	
	listensock = socket(AF_INET, SOCK_STREAM, 0);
	if (listensock < 0) {
		perror("IPv4 socket() failed");
	}

	i = 1;
	setsockopt (listensock, SOL_SOCKET, SO_REUSEADDR, &i, sizeof (int));

	if (bind(listensock, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
		perror("IPv4 bind() failed");
	}
	listen(listensock, 10);

	pthread_attr_init(&attr);
	pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN + 65536);
	
	printf("SRCP server running\n");

	if (listensock6 > listensock)
		nfds = listensock6 + 1;
	else
		nfds = listensock + 1;	

	while (1) {
		int *acceptsock;
		fd_set readset;
		socklen_t len = sizeof(from);
		socklen_t len6 = sizeof(from6);
		
		acceptsock = malloc(sizeof(int));
		
		FD_ZERO(&readset);
		FD_SET(listensock, &readset);
		FD_SET(listensock6, &readset);
		
		select(nfds, &readset, NULL, NULL, NULL);
		
		if (FD_ISSET(listensock, &readset))
			*acceptsock = accept(listensock, (struct sockaddr *)&from, &len);
		else if (FD_ISSET(listensock6, &readset))
			*acceptsock = accept(listensock6, (struct sockaddr *)&from6, &len6);
		else {
			free(acceptsock);
			continue;
		}
			
		i = 1;
		setsockopt (*acceptsock, SOL_SOCKET, SO_REUSEADDR, &i, sizeof (int));
		i = 1;
		setsockopt (listensock, IPPROTO_TCP, TCP_NODELAY, &i, sizeof (int));

		if (*acceptsock >= 0) {
			printf("New SRCP client\n");
			pthread_create(&thread, &attr, dcc_srcpd_client,
			    acceptsock);
		}
	}
	
	return NULL;
}

